/***************************************************************************
 *   Copyright (C) 2006 by Gerd v. Egidy                                   *
 *   gve@intra2net.com                                                     *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License version   *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include <sstream>
#include <iostream>

#include "socket_handler.hxx"
#include "t2n_exception.hxx"
#include "log.hxx"

using namespace std;

namespace libt2n
{

/// set options like fast reuse and keepalive every socket should have
void socket_handler::set_socket_options(int sock)
{
    int i=1;

    /* fast reuse enable */
    if (setsockopt(sock,SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"error setting socket option: " << strerror(errno));

    /* keepalive enable */
    if (setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE, &i, sizeof(i)) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"error setting socket option: " << strerror(errno));

    /* close on exec */
    int fdflags;
    fdflags=fcntl(sock,F_GETFD, 0);
    if (fdflags < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));

    fdflags |= FD_CLOEXEC;
    if (fcntl(sock,F_SETFD,fdflags) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));

    /* non-blocking mode */
    int flflags;
    flflags=fcntl(sock,F_GETFL,0);
    if (flflags < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));

    flflags |= O_NONBLOCK;
    if (fcntl(sock,F_SETFL,flflags) < 0)
        EXCEPTIONSTREAM(error,t2n_communication_error,"fcntl error on socket: " << strerror(errno));
}

/// close the underlying socket connection. Don't call directly, use the version provided
/// by the connection class you are using.
void socket_handler::close()
{
    // graceful shutdown
    shutdown(sock,SHUT_RDWR);
    ::close(sock);
}

/// is the underlying socket connection still open?
bool socket_handler::is_closed()
{
    int r=fcntl(sock,F_GETFL);

    return !(r & O_ACCMODE);
}

/** @brief check if new data is waiting on the raw socket
    @param[in,out] usec_timeout wait until new data is found, max timeout usecs.
            -1: wait endless
            0: return instantly
*/
bool socket_handler::data_waiting(long long usec_timeout,long long* usec_timeout_remaining)
{
    // just our socket
    fd_set active_fd_set;
    FD_ZERO (&active_fd_set);
    FD_SET (sock, &active_fd_set);

    /* set timeout */
    struct timeval tval;
    struct timeval *timeout_ptr;

    if (usec_timeout == -1)
        timeout_ptr = NULL;
    else
    {
        timeout_ptr = &tval;

        // convert timeout from long long usec to int sec + int usec
        tval.tv_sec = usec_timeout / 1000000;
        tval.tv_usec = usec_timeout % 1000000;
    }

    int ret=select (FD_SETSIZE, &active_fd_set, NULL, NULL, timeout_ptr);

    // return the timeout we did not use
    // todo: this is linux specific according to man 2 select
    if (usec_timeout > 0 && usec_timeout_remaining != NULL)
        *usec_timeout_remaining=(tval.tv_sec*1000000)+tval.tv_usec;

    if (ret > 0)
        return true;
    else
        return false;
}

/** @brief read data from the raw socket and copy it into the provided buffer
    @param buffer the buffer where to append the new data
    @param[in,out] usec_timeout wait until new data is found, max timeout usecs.
            -1: wait endless
            0: return instantly
*/
bool socket_handler::fill_buffer(std::string& buffer, long long usec_timeout, long long *timeout_remaining)
{
    // fast path for timeout==0
    if (usec_timeout==0 || data_waiting(usec_timeout,timeout_remaining))
        return fill_buffer(buffer);
    else
        return false;
}

/** @brief read data from the raw socket and copy it into the provided buffer. Returns
           instantly if no data is waiting.
    @param buffer the buffer where to append the new data
*/
bool socket_handler::fill_buffer(std::string& buffer)
{
    bool try_again=false;

    char socket_buffer[recv_buffer_size];

    int nbytes = read (sock, socket_buffer, recv_buffer_size);
    if (nbytes < 0)
    {
        if (errno == EAGAIN)
            return false;                // no data was waiting
        else if (errno == EINTR)
        {
            // interrupted, try again
            LOGSTREAM(debug,"EINTR received on read(), trying again");
            try_again=true;
        }
        else
        {
            LOGSTREAM(error,"error reading from socket : " << strerror(errno));
            // TODO: exception?
            return false;
        }
    }

    // End-of-file
    if (nbytes == 0 && !try_again)
    {
        LOGSTREAM(debug,"0 bytes received on read(), closing connection");
        close();
        return false;
    }

    // Data read -> store it
    if (nbytes > 0)
    {
        buffer.append(socket_buffer,nbytes);
        LOGSTREAM(debug,nbytes << " bytes read");
    }

    // more data waiting -> recurse
    if (data_waiting(0))
        fill_buffer(buffer);

    if (nbytes > 0)
        return true;
    else
        return false;
}

/// writes raw data to the socket. Don't use directly, use the write() function provided by the 
/// connection because it encapsulates the data.
void socket_handler::socket_write(const std::string& data)
{
    int offset = 0;
    while (offset < data.size())
    {
        unsigned int write_size=write_block_size;

        if (offset+write_size > data.size())
            write_size = data.size()-offset;

        int rtn;
        while ((rtn=::write(sock, data.data()+offset, write_size)) &&
               rtn == -1 && (errno == EAGAIN || errno == EINTR))
        {
            usleep (80000);
            LOGSTREAM(debug,"resuming write() call after EAGAIN or EINTR");
        }

        if (rtn == -1)
        {
            LOGSTREAM(error,"write() returned " << strerror(errno));
            // TODO: exception?
            return;
        }
        else if (rtn != write_size)
        {
            LOGSTREAM(error,"write() wrote " << rtn << " bytes, should have been " 
                << write_size << " (complete: " << data.size() << ")");

            // TODO: exception?
            return;
        }

        offset += write_size;
    }

    LOGSTREAM(debug,"wrote " << data.size() << " bytes");

    return;
}

}
